/*****************************************************************************
* "Pnginfo"								     *
*									     *
* Written by:  Martin Edlman			Ver 1.0.1, Sep. 1999	     *
******************************************************************************
* Program to display information (size and comments) in PNG file.	     *
* uses libpng on Linux. Not tried on other systems.			     *
* Options:								     *
* -h : on-line help.							     *
* -f : format output string						     *
*      \h - height							     *
*      \w - width							     *
*      \c[keyword] - comment						     *
*      \C[keyword] - keyword: comment 					     *
*      \k          - keywords                                                *
*      \f - file name							     *
*      Example: -f 'Image \f:\n Size: \wx\h\n Comment: \c\n'		     *
******************************************************************************
* History:								     *
* 14 Sep 99 - Version 1.0.1 by Martin Edlman.				     *
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <png.h>

#define DEFAULTFORMAT   "Size: \\wx\\h\\nComments:\\n\\C\\C"
#define PNG_BYTES_TO_CHECK 4

#define DPRINT(str)    printf("DEBUG: %s\n",str)

int num_text = 0;
char **keywords, **comments;

void usage(const char *msg) {
    printf("pnginfo: Display information from png file\n");
    printf("Usage: pnginfo [-f Format] [-h] filename\n \
Format: quoted text string used for formating of information\n \
Special characters for various information can be used\n \
  \\f            - filename of PNG file\n \
  \\c[KEYWORD]   - comment\n \
  \\C[KEYWORD]   - comment with keyword (keyword: comment)\n \
  \\k            - comment keywords\n \
  \\w            - width of image\n \
  \\h            - height of image\n \
  \\n            - new line\n \
  \\\\            - backslash\n \
\n \
You can specify keyword with \\c and \\C which prints only specified comment.\n \
Keyword is terminated by any special character. If it is the same as opening\n \
one it just terminates keyword and is ignored otherwise it's proccessed.\n \
If you use \\c or \\C without keyword all comments will be printed separated\n \
by new line.\n\
Default format will be used if not specified with -f:\n");
  printf("%s\n\n", DEFAULTFORMAT);
  printf("%s\n", msg);
  exit(1);
}

char *get_comment(char **format, char type) {
    char *comment = NULL, *result=NULL, **ks, **cs;
    int i;
    int end = 0;

    comment = (char*) malloc(1);    strcpy(comment, "\0");
    result  = (char*) malloc(1);    strcpy(result, "\0");
    while(*format[0] != '\0' && ! end) {
        end = *format[0] == '\\';
        if(end) {
            (*format)++;
            if(*format[0] == type) (*format)++;
            else                   (*format)--;
        }
        else {
            comment = (char*)realloc(comment, strlen(comment) + 2);
            sprintf(comment, "%s%c", comment, *format[0]);
            (*format)++;
        }
    }
    ks = keywords;
    cs = comments;
    if(strlen(comment) == 0) {   // show all comments
        for(i = 0; i < num_text; i++) {
            if(type == 'C') {
                result = (char*)realloc(result, strlen(result) + strlen(*ks) + 4);
                sprintf(result, "%s%s: ", result, *ks++);
            }
            result = (char*)realloc(result, strlen(result) + strlen(*cs) + 1);
            sprintf(result, "%s%s\n", result, *cs++);
        }
    }
    else {                       // show only specified comment
        i = 0;
        while(i++ < num_text && strcmp(*ks++, comment)) *cs++;
        if(i > num_text) return result;
        if(type == 'C') {
            result = (char*)realloc(result, strlen(result) + strlen(*(--ks)) + 4);
            sprintf(result, "%s%s: ", result, *ks);
        }
        result = (char*)realloc(result, strlen(result) + strlen(*cs) + 1);
        sprintf(result, "%s%s", result, *cs);
    }
    return result;
}

int main(int argc, char *argv[]) {
    int	i, size, width, height, extcode;
    char *filename = NULL, *format = NULL;
    char **cs, **ks;
    char png_head[PNG_BYTES_TO_CHECK];
    FILE *fp;
    png_structp png_ptr;
    png_infop   info_ptr;
    png_infop   end_info;
    png_bytep   row_pointer;
    int         num_text_beg = 0, num_text_end = 0;
    png_textp   text_ptr_beg = NULL, text_ptr_end = NULL;

    if(argc >= 2) {
        for(i = 1; i < argc; i++) {
            if(argv[i][0] == '-') {
                switch(argv[i][1]) {
                case 'h':  usage(""); break;
                case 'f':  if(strlen(argv[i]) > 2) {
                               format = &argv[i][2];
                           }
                           else if(argc > i+1) {
                               format = argv[++i];
                           }
                           else {
                               usage("Error: Format not specified");
                           }
                           break;
                default:   break;
                }
            }
            else {
                filename = argv[i];
            }
        }
    }

    if(format == NULL) {
        format = malloc(strlen(DEFAULTFORMAT)+1);
        strcpy(format, DEFAULTFORMAT);
    }

    if (filename) {
	if (!(fp = fopen(filename, "rb")))
            usage("Error: cannot open file");
    }
    else {
	if (!(fp = fdopen(0, "rb")))
            usage("Error: cannot read from stdin");
    }

    if(fread(png_head, 1, PNG_BYTES_TO_CHECK, fp) != PNG_BYTES_TO_CHECK) {
        fclose(fp);
        usage("Error: cannot read PNG head from file");
    }

    if(png_sig_cmp(png_head, (png_size_t)0, PNG_BYTES_TO_CHECK)) {
        fclose(fp);
        usage("Error: file doesn's seem to contain valid PNG image");
    }

    if(!(png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
                                          NULL, NULL, NULL))) {
        fclose(fp);
        usage("Error: error creating PNG structure");
    }

    if(!(info_ptr = png_create_info_struct(png_ptr))) {
        png_destroy_read_struct(&png_ptr,
                                (png_infopp)NULL,
                                (png_infopp)NULL);
        fclose(fp);
        usage("Error: error creating PNG info");
    }

    if(!(end_info = png_create_info_struct(png_ptr))) {
        png_destroy_read_struct(&png_ptr, &info_ptr,
                                (png_infopp)NULL);
        fclose(fp);
        usage("Error: error creating PNG end info");
    }

    if(setjmp(png_ptr->jmpbuf)) {
        png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
        fclose(fp);
        usage("Error: error setting jump back to me");
    }

    png_init_io(png_ptr, fp);
    png_set_sig_bytes(png_ptr, PNG_BYTES_TO_CHECK);

    png_read_info(png_ptr, info_ptr);
    width  = png_get_image_width(png_ptr, info_ptr);
    height = png_get_image_height(png_ptr, info_ptr);
    row_pointer = malloc(width + 1);
    for(i = 0; i < height; i++)
        png_read_row(png_ptr, row_pointer, NULL);
    png_read_end(png_ptr, end_info);

    png_get_text(png_ptr, info_ptr, &text_ptr_beg, &num_text_beg);
    png_get_text(png_ptr, end_info, &text_ptr_end, &num_text_end);
    num_text = num_text_beg + num_text_end;
    ks = keywords = (char**)malloc((num_text_beg + num_text_end) * sizeof(char*));
    cs = comments = (char**)malloc((num_text_beg + num_text_end) * sizeof(char*));
    for(i = 0; i < num_text_beg; i++) {
        *ks = (char*)malloc(strlen(text_ptr_beg[i].key) + 1);
        strcpy(*ks++, text_ptr_beg[i].key);
        *cs = malloc(strlen(text_ptr_beg[i].text) + 1);
        strcpy(*cs++, text_ptr_beg[i].text);
    }
    for(i = 0; i < num_text_end; i++) {
        *ks = (char*)malloc(strlen(text_ptr_beg[i].key) + 1);
        strcat(*ks++, text_ptr_end[i].key);
        *cs = malloc(strlen(text_ptr_beg[i].text) + 1);
        strcat(*cs++, text_ptr_end[i].text);
    }
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);

    while(format[0] != '\0') {
        if(format[0] == '\\') {
            format++;
            switch(format[0]) {
            case 'w': printf("%i",  width);     break;
            case 'h': printf("%i",  height);    break;
            case 'c': format++;  // skip 'c' to pass plain text
                      printf("%s",  get_comment(&format, 'c'));
                      format--;  // go back to cheat format++ at while end
                      break;
            case 'C': format++;  // skip 'C' to pass plain text
                      printf("%s",  get_comment(&format, 'C'));
                      format--;  // go back to cheat format++ at while end
                      break;
            case 'k':
                ks = keywords;
                for(i = 0; i < num_text; i++)
                    printf("%s\n", *ks++);
                break;
            case 'f': printf("%s", filename);  break;
            case 'n': printf("\n");             break;
            default:  printf("%c",  format[0]); break;
            }
         }
         else {
            printf("%c", format[0]);
         }
         format++;
    }

    fclose(fp);

}
